import tensorflow as tf


class Prediction(object):
    def __init__(self, model_path, input_tensors_names, input_fields, output_tensors_names, output_fields):
        super(Prediction, self).__init__()

        tf_conf = tf.ConfigProto()
        tf_conf.gpu_options.allow_growth = True
        self.sess = tf.Session(config=tf_conf)

        _ = tf.saved_model.loader.load(self.sess, [tf.saved_model.tag_constants.SERVING], model_path)

        self.graph = tf.get_default_graph()

        self.input_tensors = []

        for input_name in input_tensors_names:
            self.input_tensors.append(self.graph.get_tensor_by_name(input_name))

        self.input_fields = input_fields

        self.output_tensors = []

        for output_name in output_tensors_names:
            self.output_tensors.append(self.graph.get_tensor_by_name(output_name))

        self.output_fields = output_fields

        assert len(self.input_tensors) == len(self.input_fields)
        assert len(self.output_tensors) == len(self.output_fields)

    def __call__(self, data):
        feed_dict = {}

        for input_tensor, input_field in zip(self.input_tensors, self.input_fields):
            feed_dict[input_tensor] = data[input_field]

        outputs = self.sess.run(self.output_tensors, feed_dict=feed_dict)

        for output, output_field in zip(outputs, self.output_fields):
            data[output_field] = output

        return data



import json
import copy
from klein import Klein
from twisted.internet.defer import inlineCallbacks, returnValue
from twisted.internet import reactor
from twisted.internet import threads
import time
import click

from tinydb import TinyDB, Query


timeout = 3600  # seconds

db_service_endpoints = []

app = Klein()


@click.group()
def cli():
    pass


def screen_announcement_json(json_data):
    status = 0
    error = ''

    # message is in the format:
    # 'api_key' -- is the API key assigned to the host of the DL service

    # 'prediction_url' -- is the URL of the predictions service
    # 'interface_url' -- is the URL of the interface specification service
    # 'name' -- service name

    # 'modality' -- is the modality
    # 'anatomy' -- is the anatomy
    # 'task' -- is the task
    # 'description' -- is a short textual description of the service

    try:
        json_data['api_key']
    except KeyError:
        status += 1
        error += 'No api_key specified '

    try:
        json_data['prediction_url']
    except KeyError:
        status += 1
        error += 'No endpoint prediction url specified '

    try:
        json_data['interface_url']
    except KeyError:
        status += 1
        error += 'No endpoint interface url specified '

    try:
        json_data['name']
    except KeyError:
        status += 1
        error += 'No endpoint name specified '

    try:
        json_data['modality']
    except KeyError:
        status += 1
        error += 'No endpoint modality specified '

    try:
        json_data['anatomy']
    except KeyError:
        status += 1
        error += 'No endpoint anatomy specified '

    try:
        json_data['task']
    except KeyError:
        status += 1
        error += 'No endpoint task specified '

    try:
        json_data['description']
    except KeyError:
        status += 1
        error += 'No endpoint description specified '

    return status, error


def open_db_api_keys(file):
    global db_api_keys
    db_api_keys = TinyDB(file)


def announce_handler(json_data):
    status, error = screen_announcement_json(json_data)

    if status == 0:
        print 'CHECKING API KEY'
        service_api_key = json_data['api_key']

        API_KEY = Query()
        api_key_list = db_api_keys.search(API_KEY.api_key == service_api_key)

        if (api_key_list is None) or (len(api_key_list) == 0):
            status += 1
            error += 'Your API KEY has not been recognized. Contact the administrator of the announcement service '
            print error

    if status == 0:
        print 'CREATING SERVICE'
        creation_time = time.time()

        idx_delete = None

        for i in range(len(db_service_endpoints)):
            # if the developer already has a service replace old service with new
            if db_service_endpoints[i]['api_key'] == json_data['api_key']:
                idx_delete = i

        if idx_delete is not None:
            del db_service_endpoints[idx_delete]

        json_data['creation_time'] = creation_time

        db_service_endpoints.append(json_data)

    message = {
        'status': status,
        'error': error,
    }

    message = json.dumps(message)

    return message


def discover_handler():
    print 'SENDING CURRENT ENDPOINTS'

    endpoint_list = []

    for element in db_service_endpoints:
        current_time = time.time()
        if (current_time - element['creation_time']) < timeout:
            element = copy.deepcopy(element)
            element['SID'] = element['api_key'][0:7]
            element['api_key'] = ''

            endpoint_list.append(element)

    return json.dumps(endpoint_list)


@app.route('/announce', methods=['POST'])
@inlineCallbacks
def announce(request):
    data = json.loads(request.content.read())
    result = yield threads.deferToThread(announce_handler, data)

    returnValue(result)


@app.route('/discover')
@inlineCallbacks
def discover(request):
    request.setHeader('Access-Control-Allow-Origin', '*')
    request.setHeader('Access-Control-Allow-Methods', 'GET')
    request.setHeader('Access-Control-Allow-Headers', 'x-prototype-version,x-requested-with')
    request.setHeader('Access-Control-Max-Age', 2520)  # 42 hours

    result = yield threads.deferToThread(discover_handler)

    returnValue(result)


@click.command()
@click.option('--db_filename', default='./db/api_key_db.json')
@click.option('--port', default=8000)
def start_service(db_filename, port):
    open_db_api_keys(db_filename)
    app.run(port=port, host='0.0.0.0')
    reactor.run()


cli.add_command(start_service)


if __name__ == '__main__':
    cli()





import torch
import torch.backends.cudnn as cudnn
import numpy as np


class Prediction(object):
    def __init__(self, model_path, input_arg_names, input_fields, output_fields, with_gpu=True):
        super(Prediction, self).__init__()

        self.model = torch.load(model_path)
        self.with_gpu = with_gpu

        if self.with_gpu:
            self.model.cuda()
            # avoid nonsense from cudnn
            cudnn.enabled = True
            cudnn.benchmark = True

        self.input_fields = input_fields
        self.input_arg_names = input_arg_names

        assert len(self.input_fields) == len(self.input_arg_names)

        self.output_fields = output_fields

    def __call__(self, data):
        arg_dict = {}

        for arg_name, field_name in zip(self.input_arg_names, self.input_fields):
            arg_dict[arg_name] = torch.from_numpy(data[field_name])

            if self.with_gpu:
                arg_dict[arg_name].cuda()

        outputs = self.model(**arg_dict)

        if not hasattr(outputs, 'len'):
            outputs = [outputs]

        assert len(outputs) == len(self.output_fields)

        for output, output_field in zip(outputs, self.output_fields):
            data[output_field] = output.cpu().detach().numpy().astype(dtype=np.float32)

        return data


import click
import json
import os
import nibabel as nib
import SimpleITK as sitk
import numpy as np

from tomaat.server import TomaatServiceDelayedResponse


# this extra works with model zoo from NiftyNet and in particular with highresnet3d_brain_parcellation model.


input_interface = \
    [
        {'type': 'volume', 'destination': 'T1'},
        {'type': 'checkbox', 'destination': 'alt_convention', 'text': 'use alternative coordinate convention'},
    ]

output_interface = \
    [
        {'type': 'LabelVolume', 'field': 'labels'}
    ]


class FromNumpyToSITK(object):
    def __init__(self,
                 fields,
                 field_original_spacing='original_spacings_NP',
                 field_original_direction='original_directions_NP',
                 field_original_origins='original_origins_NP'
                 ):
        '''
        FromNumpyToSITK converts numpy array data in SITK data
        :param fields: fields of the dictionary whose content should be modified
        :param field_original_spacing: field data dictionary used to store original spacing of sitk volumes
        :param field_original_direction: field data dictionary used to store original direction of sitk volumes
        :param field_original_origins: field data dictionary used to store original origin coordinate of sitk volumes
        '''
        super(FromNumpyToSITK, self).__init__()
        self.fields = fields
        self.field_original_spacing = field_original_spacing
        self.field_original_direction = field_original_direction
        self.field_original_origins = field_original_origins

    def __call__(self, data):
        original_spacings = data[self.field_original_spacing]
        original_directions = data[self.field_original_direction]
        original_origins = data[self.field_original_origins]
        for field in self.fields:
            for i in range(len(data[field])):
                numpy_data = np.transpose(data[field][i], [2, 1, 0])

                print(original_spacings[field][i].tolist())

                data[field][i] = sitk.GetImageFromArray(numpy_data)
                data[field][i].SetDirection(original_directions[field][i])
                data[field][i].SetOrigin(original_origins[field][i])
                data[field][i].SetSpacing(original_spacings[field][i].tolist())

        return data


class FromNiftyToNumpy(object):
    def __init__(self, field):
        self.field = field

    def __call__(self, data):
        entries = []
        spacings = []
        origins = []
        directions = []

        for entry, convention in zip(data[self.field], data['alt_convention']):
            nii_image = nib.load(entry)

            spacing = np.asarray([1., 1., 1.])
            affine = nii_image.header.get_sform()

            spacings.append(spacing)
            if convention == 'True':
                directions.append((affine[0:3, 0:3]).flatten().tolist())
                origins.append(affine[0:3, 3].flatten().tolist())
            else:
                affine = np.matmul(np.diag([-1, -1, 1, 1]), affine)
                directions.append(affine[0:3, 0:3].flatten().tolist())
                origins.append(affine[0:3, 3].flatten().tolist())

            numpy_image = np.squeeze(nii_image.get_data().astype(np.float32))

            entries.append(numpy_image)

        data[self.field] = entries
        data['spacings'] = {self.field: spacings}
        data['origins'] = {self.field: origins}
        data['directions'] = {self.field: directions}


        return data


class NiftyNetZooApp(object):
    def __init__(self, ini_file):
        self.ini_file = ini_file

    def __call__(self, data, gpu_lock):
        # lock mechanism

        with open(self.ini_file, 'r') as f:
            txt = f.read()

        dirname = os.path.dirname(data['T1'][0])

        txt = txt.replace('<dir>', dirname)

        new_conf = os.path.join(dirname, './default.ini')

        with open(new_conf, 'w+') as f:
            f.write(txt)

        command = 'net_segment inference -c {}'.format(new_conf)

        if gpu_lock is not None:
            gpu_lock.acquire()  # acquire GPU lock

        os.system(command)

        if gpu_lock is not None:
            gpu_lock.release()  # release GPU lock

        transform_1 = FromNiftyToNumpy(field='labels')
        transform_2 = FromNumpyToSITK(
            fields=['labels'],
            field_original_direction='directions',
            field_original_spacing='spacings',
            field_original_origins='origins',
        )

        img_basename = os.path.splitext(data['T1'][0])[0].replace("-", "")

        label_name = img_basename + '_niftynet_out.nii.gz'

        data['labels'] = [img_basename + '_niftynet_out.nii.gz']

        data = transform_2(transform_1(data))

        os.remove(new_conf)
        os.remove(data['T1'][0])
        os.remove(label_name)

        return data


@click.group()
def cli():
    pass


@click.command()
@click.option('--config_file_path')
@click.option('--ini_file_path')
def start_service(config_file_path, ini_file_path):
    """
    :type config_file_path: str valid path of the config file in JSON format
    :type ini_file_path: str valid path of the config ini file for the niftynet model
    :return: None
    """
    with open(config_file_path) as f:
        config = json.load(f)

    application = NiftyNetZooApp(ini_file_path)

    service = TomaatServiceDelayedResponse(
        config=config,
        app=application,
        input_interface=input_interface,
        output_interface=output_interface
    )

    if config['announce']:
        service.start_service_announcement()

    service.run()


cli.add_command(start_service)


if __name__ == '__main__':
    cli()



class TransformChain(object):
    def __init__(self, transforms_list):
        super(TransformChain, self).__init__()
        self.transforms_list = transforms_list

    def __call__(self, data):
        for transform in self.transforms_list:
            data = transform(data)

        return data



import click
import os

from tinydb import TinyDB, Query
from binascii import hexlify


key_length = 64


@click.group()
def cli():
    pass


@click.command()
@click.option('--api_key')
@click.option('--db_filename', default='./db/api_key_db.json')
@click.option('--remove', default=False)
def add_db_record(api_key, db_filename, remove):
    db = TinyDB(db_filename)
    if len(api_key) == key_length:
        if not remove:
            db.insert({'api_key': api_key})
        else:
            API_KEY = Query()
            el = db.get(API_KEY.api_key == api_key)
            if el is not None:
                db.remove(doc_ids=[el.doc_id])
    db.close()


@click.command()
def create_api_key():
    print hexlify(os.urandom(key_length/2))


cli.add_command(add_db_record)
cli.add_command(create_api_key)


if __name__ == '__main__':
    cli()


import json
import requests
import tempfile
import uuid
import os
import SimpleITK as sitk
import base64
import traceback
import numpy as np
import shutil

try:
    # For Python 3.0 and later
    from urllib.request import urlopen
except ImportError:
    # Fall back to Python 2's urllib2
    from urllib2 import urlopen

from multiprocessing import Process, Manager, Lock
from urllib2 import urlopen

from klein import Klein
from twisted.internet.defer import inlineCallbacks, returnValue, DeferredLock
from twisted.internet import threads
from twisted.internet.task import LoopingCall
from twisted.internet import reactor
from twisted.logger import Logger


ANNOUNCEMENT_SERVER_URL = 'http://tomaat.cloud:8001/announce'
ANNOUNCEMENT_INTERVAL = 1600  # seconds

logger = Logger()


def do_announcement(announcement_server_url, message):
    json_message = json.dumps(message)

    try:
        response = requests.post(announcement_server_url, data=json_message)

        response_json = response.json()

        if response_json['status'] != 0:
            logger.error('status {}'.format(response_json['status']))
            logger.error('errors: {}'.format(response_json['error']))
    except:
        logger.error('WARNING: ERROR while connecting to announcement service.')
        pass


class TomaatApp(object):
    """
    A TomaatApp is an object that implements the functionality of the user application. More specifically,
    this object implements the workflow that is needed by the app. It requires the following arguments
    """
    def __init__(self, preprocess_fun, inference_fun, postprocess_fun):
        """
        To instantiate a TomaatApp the following arguments are needed
        :type preprocess_fun: Callable function or callable object implementing pre-processing
        :type inference_fun: Callable function or callable object implementing inference
        :type postprocess_fun: Callable function or callable object implementing post-processing
        """
        super(TomaatApp, self).__init__()
        self.preprocess_fun = preprocess_fun
        self.inference_fun = inference_fun
        self.postprocess_fun = postprocess_fun

    def __call__(self, data, gpu_lock=None):
        """
        When a TomaatApp object is called it performs pre-processing, inference and post-processing
        :type data: dict dictionary containing data. The dictionary must contain the fields expected by pre-processing
        :type gpu_lock: DeferredLock optional lock to allow threads to safely use the GPU. No GPU => no lock needed
        :return: dict containing inference results after post-processing
        """
        transformed_data = self.preprocess_fun(data)

        if gpu_lock is not None:
            gpu_lock.acquire()  # acquire GPU lock

        result = self.inference_fun(transformed_data)  # GPU call

        if gpu_lock is not None:
            gpu_lock.release()  # release GPU lock

        transformed_result = self.postprocess_fun(result)

        return transformed_result


class TomaatService(object):
    klein_app = Klein()

    announcement_task = None

    gpu_lock = DeferredLock()

    def __init__(self, config, app, input_interface, output_interface):
        """
        To instantitate a TomaatService the following arguments are needed
        :type config: dict dictionary containing configuration for the service
        :type app: TomaatApp implementing the user application to be served through TomaatService
        :type input_interface: dict containing the specification for input interface (See documentation or example)
        :type output_interface: dict containing the specification for output interface (See documentation or example)
        """
        super(TomaatService, self).__init__()

        self.config = config
        self.app = app

        self.input_interface = input_interface
        self.output_interface = output_interface

    @klein_app.route('/interface', methods=['GET'])
    def interface(self, _):
        return json.dumps(self.input_interface)

    @klein_app.route('/predict', methods=['POST'])
    @inlineCallbacks
    def predict(self, request):
        logger.info('predicting...')

        result = yield threads.deferToThread(self.received_data_handler, request)

        returnValue(result)

    def start_service_announcement(
            self,
            fun=do_announcement,
            announcement_server_url=ANNOUNCEMENT_SERVER_URL,
            delay=ANNOUNCEMENT_INTERVAL
    ):
        try:
            api_key = self.config['api_key']
        except KeyError:
            raise ValueError('Api-key is missing')

        try:
            host = self.config['host']
        except KeyError:
            ip = urlopen('http://ip.42.pl/raw').read()
            port = self.config['port']
            host = 'http://' + str(ip) + ':' + str(port) + '/'
            pass

        message = {
            'api_key': api_key,
            'prediction_url': host+'/predict',
            'interface_url': host+'/interface',
            'name': self.config['name'],
            'modality': self.config['modality'],
            'task': self.config['task'],
            'anatomy': self.config['anatomy'],
            'description': self.config['description'],
        }

        self.announcement_task = LoopingCall(fun, *(announcement_server_url, message))

        self.announcement_task.start(delay)

    def stop_service_announcement(self):
        self.announcement_task.stop()

    def make_error_response(self, message):
        """
        Create simple error message to be returned to the client as plain text
        :type message: str error message to be returned to the client
        :return: response to be returned to client
        """
        response = [{'type': 'PlainText', 'content': message, 'label': 'Error!'}]
        return response

    def parse_request(self, request, savepath):
        """
        This function takes in the content of the client message and creates a dictionary containing data.
        The service interface, that was specified in the input_interface dictionary specified at init,
        contains the specifications of the data that is needed to run this service and the fields of the dictionary
        returned by this function where the client data should be stored.
        :type request: dict request sent by the client
        :return: dict containing data that can be fed to the pre-processing, inference, post-processing pipeline
        """

        data = {}

        for element in self.input_interface:
            raw = request.args[element['destination']]

            if element['type'] == 'volume':
                uid = uuid.uuid4()

                mha_file = str(uid).replace('-', '') + '.mha'

                tmp_filename_mha = os.path.join(savepath, mha_file)

                with open(tmp_filename_mha, 'wb') as f:
                    f.write(raw[0])

                data[element['destination']] = [tmp_filename_mha]

            elif element['type'] == 'slider':
                data[element['destination']] = [float(raw[0])]

            elif element['type'] == 'checkbox':
                data[element['destination']] = [str(raw[0])]

            elif element['type'] == 'radiobutton':
                data[element['destination']] = [str(raw[0])]

            elif element['type'] == 'fiducials':
                # Each coordinate is separated by ';'
                # Each coord value is separated by ','
                fiducial_string = str(raw[0])
                fiducial_list = [ [ float(val) for val in coords.split(',')] for coords in fiducial_string.split(';')]
                data[element['destination']] = [np.asarray(fiducial_list)]

        return data

    def make_response(self, data, savepath):
        """
        This function takes in the post-processed results of inference and creates a message for the client.
        The message is created according to the directives specified in the output_interface dictionary passed
        during instantiation of TomaatService object.
        :type request: dict containing the inference results (stored in the appropriate fields)
        :return: JSON containing response that can be returned to the client
        """
        message = []

        for element in self.output_interface:
            type = element['type']
            field = element['field']

            if type == 'LabelVolume':
                uid = uuid.uuid4()

                mha_seg = str(uid).replace('-', '') + '_seg.mha'
                tmp_label_volume = os.path.join(savepath, mha_seg)

                writer = sitk.ImageFileWriter()
                writer.SetFileName(tmp_label_volume)
                writer.SetUseCompression(True)
                writer.Execute(data[field][0])

                with open(tmp_label_volume, 'rb') as f:
                    vol_string = base64.encodestring(f.read())

                message.append({'type': 'LabelVolume', 'content': vol_string, 'label': ''})

                os.remove(tmp_label_volume)

            elif type == 'VTKMesh':
                import vtk

                uid = uuid.uuid4()

                vtk_mesh = str(uid).replace('-', '') + '_seg.vtk'
                tmp_vtk_mesh = os.path.join(savepath, vtk_mesh)

                writer = vtk.vtkPolyDataWriter()
                writer.SetFileName(tmp_vtk_mesh)
                writer.SetInput(data[field][0])
                writer.SetFileTypeToASCII()
                writer.Write()

                with open(tmp_vtk_mesh, 'rb') as f:
                    mesh_string = base64.encodestring(f.read())

                message.append({'type': 'VTKMesh', 'content': mesh_string, 'label': ''})

                os.remove(tmp_vtk_mesh)

            elif type == 'PlainText':
                message.append({'type': 'PlainText', 'content': str(data[field][0]), 'label': ''})

            elif type == 'Fiducials':
                fiducial_array = data[field][0]
                fiducial_str = ';'.join([','.join(map(str,fid_point)) for fid_point in fiducial_array])
                message.append({'type': 'Fiducials', 'content': fiducial_str, 'label': ''})

        return message

    def received_data_handler(self, request):
        savepath = os.path.join(tempfile.gettempdir(), str(uuid.uuid4()).replace('-', ''))

        os.mkdir(savepath)

        try:
            data = self.parse_request(request, savepath)
        except:
            traceback.print_exc()
            logger.error('Server-side ERROR during request parsing')
            response = self.make_error_response('Server-side ERROR during request parsing')
            return json.dumps(response)

        try:
            transformed_result = self.app(data, gpu_lock=self.gpu_lock)
        except:
            traceback.print_exc()
            logger.error('Server-side ERROR during processing')
            response = self.make_error_response('Server-side ERROR during processing')
            return json.dumps(response)

        try:
            response = self.make_response(transformed_result, savepath)
        except:
            traceback.print_exc()
            logger.error('Server-side ERROR during response message creation')
            response = self.make_error_response('Server-side ERROR during response message creation')
            return json.dumps(response)

        shutil.rmtree(savepath)
        
        return json.dumps(response)

    def run(self):
        self.klein_app.run(port=self.config['port'], host='0.0.0.0')
        reactor.run()


class TomaatServiceDelayedResponse(TomaatService):
    announcement_task = None

    gpu_lock = DeferredLock()
    
    multiprocess_manager = Manager()

    result_dict = multiprocess_manager.dict()
    reqest_list = multiprocess_manager.list()

    multiprocess_lock = Lock()

    klein_app = Klein()

    def __init__(self, no_concurrent_thread_execution=True, **kwargs):
        super(TomaatServiceDelayedResponse, self).__init__(**kwargs)
        self.no_concurrent_thread_execution = no_concurrent_thread_execution

    def received_data_handler(self, request):
        req_id = str(uuid.uuid4()).replace('-', '')

        savepath = os.path.join(tempfile.gettempdir(), req_id)

        os.mkdir(savepath)

        def processing_thread():
            if self.no_concurrent_thread_execution:
                self.multiprocess_lock.acquire()

            response = self.make_error_response('Server-side ERROR during processing')

            try:
                data = self.parse_request(request, savepath)
            except:
                traceback.print_exc()
                logger.error('Server-side ERROR during request parsing')

            try:
                transformed_result = self.app(data, gpu_lock=self.gpu_lock)
            except:
                traceback.print_exc()
                logger.error('Server-side ERROR during processing')

            try:
                response = self.make_response(transformed_result, savepath)
            except:
                traceback.print_exc()
                logger.error('Server-side ERROR during response message creation')

            response = [{
                'type': 'PlainText',
                'content': 'The results of your earlier request {} have been received'.format(req_id),
                'label': ''
            }] + response

            self.result_dict[req_id] = response

            shutil.rmtree(savepath)

            if self.no_concurrent_thread_execution:
                self.multiprocess_lock.release()

        delegated_process = Process(target=processing_thread, args=())
        delegated_process.start()

        self.reqest_list.append(req_id)

        response = [{'type': 'DelayedResponse', 'request_id': req_id}]

        return json.dumps(response)

    @klein_app.route('/interface', methods=['GET'])
    def interface(self, _):
        return json.dumps(self.input_interface)

    @klein_app.route('/predict', methods=['POST'])
    @inlineCallbacks
    def predict(self, request):
        logger.info('predicting...')

        result = yield threads.deferToThread(self.received_data_handler, request)

        returnValue(result)

    @klein_app.route('/responses', methods=['POST'])
    @inlineCallbacks
    def responses(self, request):
        logger.info('getting responses...')

        result = yield threads.deferToThread(self.responses_data_handler, request)

        returnValue(result)

    def responses_data_handler(self, request):
        req_id = request.args['request_id'][0]

        print(req_id)
        print(self.reqest_list)

        if req_id not in self.reqest_list:
            response = [{
                'type': 'PlainText',
                'content': 'The results of request {} cannot be retrieved'.format(req_id),
                'label': ''
            }]

            return json.dumps(response)

        try:
            response = self.result_dict[req_id]
            #removing content of list and dict
            del self.result_dict[req_id]
            self.reqest_list.remove(req_id)
        except KeyError:
            response = [{'type': 'DelayedResponse', 'request_id': req_id}]


        return json.dumps(response)


import click
import json

from tomaat.server import TomaatApp, TomaatService
from tomaat.frameworks.tf import Prediction
from tomaat.extras import TransformChain
from tomaat.extras import (
    FromSITKToNumpy,
    FromITKFormatFilenameToSITK,
    FromNumpyOriginalSizeToStandardSize,
    FromSITKOriginalIntensitiesToRescaledIntensities,
    FromListToNumpy5DArray,
    FromSITKUint8ToSITKFloat32,
    FromSITKOriginalResolutionToStandardResolution,
    ThresholdNumpy,
    FromNumpyToSITK,
    FromNumpyStandardSizeToOriginalSize,
    FromNumpy5DArrayToList,
    FromSITKStandardResolutionToOriginalResolution,
)


input_interface = \
    [
        {'type': 'volume', 'destination': 'images'},
        {'type': 'slider', 'destination': 'threshold', 'minimum': 0, 'maximum': 1},
        {'type': 'checkbox', 'destination': 'RAS', 'text': 'use slicer coordinate conventions'},
        {'type': 'radiobutton', 'destination': 'spacing_metric', 'text': 'spacing metric', 'options': ['millimeters', 'meters']},
    ]

output_interface = \
    [
        {'type': 'LabelVolume', 'field': 'images'}
    ]


def create_pre_process_pipeline(config):
    """
    This is an application-specific function that creates the pre_processing pipeline. it returns a callable.
    :type config: dict a dictionary containing the configuration for this specific task
    :return: a callable
    """
    transform_1 = FromITKFormatFilenameToSITK(fields=['images'])
    transform_2 = FromSITKUint8ToSITKFloat32(fields=['images'])
    transform_3 = FromSITKOriginalIntensitiesToRescaledIntensities(fields=['images'])
    transform_4 = FromSITKOriginalResolutionToStandardResolution(
        fields=['images'],
        resolution=config['volume_resolution'],
        field_spacing_metric='spacing_metric'
    )
    transform_5 = FromSITKToNumpy(fields=['images'])
    transform_6 = FromNumpyOriginalSizeToStandardSize(fields=['images'], size=config['volume_size'])
    transform_7 = FromListToNumpy5DArray(fields=['images'])

    pre_process_pipeline = TransformChain(
        [transform_1,
         transform_2,
         transform_3,
         transform_4,
         transform_5,
         transform_6,
         transform_7,
         ]
    )

    return pre_process_pipeline


def create_post_process_pipeline(config):
    """
    This is an application-specific function that creates the post_processing pipeline. it returns a callable.
    :type config: dict a dictionary containing the configuration for this specific task
    :return: a callable
    """
    antitransform_8 = ThresholdNumpy(image_field='images', threshold_field='threshold')
    antitransform_7 = FromNumpy5DArrayToList(fields=['images'])
    antitransform_6 = FromNumpyStandardSizeToOriginalSize(fields=['images'])
    antitransform_5 = FromNumpyToSITK(fields=['images'])
    antitransform_4 = FromSITKStandardResolutionToOriginalResolution(fields=['images'])

    post_process_pipeline = TransformChain(
        [antitransform_8,
         antitransform_7,
         antitransform_6,
         antitransform_5,
         antitransform_4,
         ]
    )

    return post_process_pipeline


@click.group()
def cli():
    pass


@click.command()
@click.option('--model_path')
@click.option('--config_file_path')
def start_service(model_path, config_file_path):
    """

    :type model_path: str valid path of the TF model where a .pb model file is present
    :type config_file_path: str valid path of the config file in JSON format
    :return: None
    """
    with open(config_file_path) as f:
        config = json.load(f)

    pre_process_pipeline = create_pre_process_pipeline(config)
    post_process_pipeline = create_post_process_pipeline(config)

    tensorflow_prediction = Prediction(
        model_path=model_path,
        input_tensors_names=["images:0"],
        input_fields=["images"],
        output_tensors_names=["logits:0"],
        output_fields=["images"]  # replace images with results because it's more convenient for the transforms
    )

    application = TomaatApp(
        preprocess_fun=pre_process_pipeline,
        inference_fun=tensorflow_prediction,
        postprocess_fun=post_process_pipeline
    )

    service = TomaatService(
        config=config,
        app=application,
        input_interface=input_interface,
        output_interface=output_interface
    )

    if config['announce']:
        service.start_service_announcement()

    service.run()


cli.add_command(start_service)


if __name__ == '__main__':
    cli()



from setuptools import setup, find_packages

setup(
    name='tomaat',
    version="0.1.3",
    description='TOMAAT SERVER SIDE',
    url='https://github.com/fmilletari/tomaat',
    author='Fausto Milletari',
    author_email='tomaat.segmentation@gmail.com',
    packages=find_packages(),
    install_requires=[
        'klein',
        'SimpleITK',
        'numpy',
        'requests',
        'click',
        'tinydb',
    ],
)



import SimpleITK as sitk
import numpy as np
import os

'''
NOTE: The transforms that are added to the data during inference must be IDENTICAL to 
those used during training on the original data, at least for what concerns the forward transform step
'''


class FromITKFormatFilenameToSITK(object):
    def __init__(self, fields):
        '''
        FromITKFormatFilenameToSITK loads ITK compatible files
        :param fields: fields of the dictionary whose content should be replaced by SITK images
        '''
        self.fields = fields

    def __call__(self, data):
        for field in self.fields:
            volume_list = []
            for elem in data[field]:
                volume = sitk.ReadImage(elem)
                volume_list.append(volume)
                os.remove(elem)
            data[field] = volume_list

        return data


class FromSITKUint8ToSITKFloat32(object):
    def __init__(self, fields):
        '''
        FromSITKUint8ToSITKFloat32 casts data to float32.
        :param fields: fields of the dictionary whose content should be modified
        '''
        super(FromSITKUint8ToSITKFloat32, self).__init__()
        self.fields = fields

    def __call__(self, data):
        filter = sitk.CastImageFilter()
        filter.SetOutputPixelType(sitk.sitkFloat32)
        for field in self.fields:
            for i in range(len(data[field])):
                data[field][i] = filter.Execute(data[field][i])

        return data


class FromSITKFloat32ToSITKUint8(object):
    def __init__(self, fields):
        '''
        FromSITKFloat32ToSITKUint8 method casts data back to uint8.
        :param fields: fields of the dictionary whose content should be modified
        '''
        super(FromSITKFloat32ToSITKUint8, self).__init__()
        self.fields = fields

    def __call__(self, data):
        filter = sitk.CastImageFilter()
        filter.SetOutputPixelType(sitk.sitkUInt8)
        for field in self.fields:
            for i in range(len(data[field])):
                data[field][i] = filter.Execute(data[field][i])

        return data


class FromSITKOriginalIntensitiesToRescaledIntensities(object):
    def __init__(self,
                 fields,
                 min_intensity=0.,
                 max_intensity=1.,
                 field_original_ranges_min='original_ranges_min',
                 field_original_ranges_max='original_ranges_max'
                 ):
        '''
        FromSITKOriginalIntensitiesToRescaledIntensities method rescales data the correct intensity range.
        :param fields: fields of the dictionary whose content should be modified
        :param min_intensity: the lower end of the new intensity range
        :param max_intensity: the higher end of the new intensity range
        :param field_original_ranges_min: field to use in data dictionary to store original intensity minima
        :param field_original_ranges_max: field to use in data dictionary to store original intensity maxima
        '''
        super(FromSITKOriginalIntensitiesToRescaledIntensities, self).__init__()

        self.fields = fields
        self.min = min_intensity
        self.max = max_intensity
        self.field_original_ranges_min = field_original_ranges_min
        self.field_original_ranges_max = field_original_ranges_max

    def __call__(self, data):
        original_ranges_min = {}
        original_ranges_max = {}

        rescaling_fiter = sitk.RescaleIntensityImageFilter()
        rescaling_fiter.SetOutputMaximum(self.max)
        rescaling_fiter.SetOutputMinimum(self.min)

        min_max_filter = sitk.MinimumMaximumImageFilter()

        for field in self.fields:
            original_ranges_max[field] = []
            original_ranges_min[field] = []
            for i in range(len(data[field])):
                min_max_filter.Execute(data[field][i])

                original_ranges_max[field].append(min_max_filter.GetMaximum())
                original_ranges_min[field].append(min_max_filter.GetMinimum())

                data[field][i] = rescaling_fiter.Execute(data[field][i], self.min, self.max)

        data[self.field_original_ranges_min] = original_ranges_min
        data[self.field_original_ranges_max] = original_ranges_max

        return data


class FromSITKRescaledIntensitiesToOriginalIntensities(object):
    def __init__(self,
                 fields,
                 field_original_ranges_min='original_ranges_min',
                 field_original_ranges_max='original_ranges_max'
                 ):
        '''
        FromSITKRescaledIntensitiesToOriginalIntensities rescales data the original intensity range.
        :param fields: fields of the dictionary whose content should be modified
        :param field_original_ranges_min: field in the data dictionary containing the minima of the original volumes
        :param field_original_ranges_max: field in the data dictionary containing the maxima of the original volumes
        '''
        super(FromSITKRescaledIntensitiesToOriginalIntensities, self).__init__()

        self.fields = fields
        self.field_original_ranges_min = field_original_ranges_min
        self.field_original_ranges_max = field_original_ranges_max

    def __call__(self, data):
        original_ranges_min = data[self.field_original_ranges_min]
        original_ranges_max = data[self.field_original_ranges_max]
        for field in self.fields:
            for i in range(len(data[field])):
                rescaling_fiter = sitk.RescaleIntensityImageFilter()

                rescaling_fiter.SetOutputMaximum(original_ranges_max[field][i])
                rescaling_fiter.SetOutputMinimum(original_ranges_min[field][i])

                data[field][i] = rescaling_fiter.Execute(data[field][i])

        return data


class FromSITKOriginalResolutionToStandardResolution(object):
    def __init__(self, fields, resolution, field_original_spacings='original_spacings', field_spacing_metric=None):
        '''
        FromSITKOriginalResolutionToStandardResolution makes all the volumes have the same resolution
        :param fields: fields of the dictionary whose content should be modified
        :param resolution: the new resolution of the data in three directions
        :param field_original_spacings: field to use for the data dictionary to store original resolutions
        '''
        super(FromSITKOriginalResolutionToStandardResolution, self).__init__()
        self.fields = fields
        self.resolution = resolution
        self.field_original_spacings = field_original_spacings
        self.field_spacing_metric = field_spacing_metric

    def __call__(self, data):
        original_spacings = {}

        for field in self.fields:
            original_spacings[field] = []

            for i in range(len(data[field])):

                if self.field_spacing_metric is not None:
                    if data[self.field_spacing_metric][i] == 'meters':
                        resolution = np.asarray(self.resolution) / 1000.
                    elif data[self.field_spacing_metric][i] == 'millimeters':
                        resolution = np.asarray(self.resolution)
                else:
                    resolution = np.asarray(self.resolution)

                factor = np.asarray(data[field][i].GetSpacing()) / np.asarray(resolution, dtype=float)
                new_size = np.asarray(data[field][i].GetSize() * factor, dtype=int)

                original_spacings[field].append(data[field][i].GetSpacing())

                resampler = sitk.ResampleImageFilter()
                resampler.SetReferenceImage(data[field][i])
                resampler.SetOutputSpacing(resolution)
                resampler.SetSize(new_size)

                data[field][i] = resampler.Execute(data[field][i])
        data[self.field_original_spacings] = original_spacings

        return data


class FromSITKStandardResolutionToOriginalResolution(object):
    def __init__(self, fields, field_original_spacings='original_spacings'):
        '''
        FromSITKStandardResolutionToOriginalResolution makes all the volumes return to their original resolution
        :param fields: fields of the dictionary whose content should be modified
        :param field_original_spacings: the fields of the data dictionary where the original resolutions are stored
        '''
        super(FromSITKStandardResolutionToOriginalResolution, self).__init__()
        self.fields = fields
        self.field_original_spacings = field_original_spacings

    def __call__(self, data):
        original_spacings = data[self.field_original_spacings]

        for field in self.fields:
            for i in range(len(data[field])):
                factor = \
                    np.asarray(data[field][i].GetSpacing()) / np.asarray(original_spacings[field][i], dtype=float)
                new_size = np.asarray(data[field][i].GetSize() * factor, dtype=int)

                resampler = sitk.ResampleImageFilter()
                resampler.SetReferenceImage(data[field][i])
                resampler.SetOutputSpacing(original_spacings[field][i])
                resampler.SetSize(new_size)

                data[field][i] = resampler.Execute(data[field][i])

        return data


class FromSITKToNumpy(object):
    def __init__(self,
                 fields,
                 field_original_spacing='original_spacings_NP',
                 field_original_direction='original_directions_NP',
                 field_original_origins='original_origins_NP'
                 ):
        '''
        FromSITKToNumpy converts SITK data in numpy arrays
        :param fields: fields of the dictionary whose content should be modified
        :param field_original_spacing: field data dictionary used to store original spacing of sitk volumes
        :param field_original_direction: field data dictionary used to store original direction of sitk volumes
        :param field_original_origins: field data dictionary used to store original origin coordinate of sitk volumes
        '''
        super(FromSITKToNumpy, self).__init__()
        self.fields = fields
        self.field_original_spacing = field_original_spacing
        self.field_original_direction = field_original_direction
        self.field_original_origins = field_original_origins

    def __call__(self, data):
        original_directions = {}
        original_origins = {}
        original_spacings = {}

        for field in self.fields:
            original_spacings[field] = []
            original_directions[field] = []
            original_origins[field] = []

            for i in range(len(data[field])):
                original_spacings[field].append(data[field][i].GetSpacing())
                original_directions[field].append(data[field][i].GetDirection())
                original_origins[field].append(data[field][i].GetOrigin())

                data[field][i] = np.transpose(sitk.GetArrayFromImage(data[field][i]).astype(dtype=np.float32), [2, 1, 0])

        data[self.field_original_spacing] = original_spacings
        data[self.field_original_direction] = original_directions
        data[self.field_original_origins] = original_origins

        return data


class FromNumpyToSITK(object):
    def __init__(self,
                 fields,
                 field_original_spacing='original_spacings_NP',
                 field_original_direction='original_directions_NP',
                 field_original_origins='original_origins_NP'
                 ):
        '''
        FromNumpyToSITK converts numpy array data in SITK data
        :param fields: fields of the dictionary whose content should be modified
        :param field_original_spacing: field data dictionary used to store original spacing of sitk volumes
        :param field_original_direction: field data dictionary used to store original direction of sitk volumes
        :param field_original_origins: field data dictionary used to store original origin coordinate of sitk volumes
        '''
        super(FromNumpyToSITK, self).__init__()
        self.fields = fields
        self.field_original_spacing = field_original_spacing
        self.field_original_direction = field_original_direction
        self.field_original_origins = field_original_origins

    def __call__(self, data):
        original_spacings = data[self.field_original_spacing]
        original_directions = data[self.field_original_direction]
        original_origins = data[self.field_original_origins]
        for field in self.fields:
            for i in range(len(data[field])):
                numpy_data = np.transpose(data[field][i], [2, 1, 0])

                data[field][i] = sitk.GetImageFromArray(numpy_data)
                data[field][i].SetDirection(original_directions[field][i])
                data[field][i].SetOrigin(original_origins[field][i])
                data[field][i].SetSpacing(original_spacings[field][i])

        return data


class FromNumpyOriginalSizeToStandardSize(object):
    def __init__(self,
                 fields,
                 size,
                 field_pads='pads_std_size',
                 field_crops="crops_std_size",
                 field_original_sizes='original_sizes_std_size'
                 ):
        '''
        FromNumpyOriginalSizeToStandardSize resizes data to predefined size.
        :param fields: fields of the dictionary whose content should be modified
        :param size: desired image size in the three directions
        :param field_pads: field of data dictionary to use to store paddings used in this transform
        :param field_crops: field of data dictionary to use to store crops used in this transform
        :param field_original_sizes: field of data dictionary to use to store sizes used in this transform
        '''
        self.fields = fields
        self.size = np.asarray(size)

        self.field_pads = field_pads
        self.field_crops = field_crops
        self.field_original_sizes = field_original_sizes

    def __call__(self, data):
        pads = {}
        crops = {}
        original_sizes = {}
        for field in self.fields:
            pads[field] = []
            crops[field] = []
            original_sizes[field] = []
            for i in range(len(data[field])):
                original_sizes[field].append(np.asarray(data[field][i].shape))

                pad_amount = self.size - data[field][i].shape

                pad_before = np.floor(pad_amount / 2.).astype(int)
                pad_after = np.ceil(pad_amount / 2.).astype(int)

                pad_before[pad_before < 0] = 0
                pad_after[pad_after < 0] = 0

                pad_vec = ((pad_before[0], pad_after[0]), (pad_before[1], pad_after[1]), (pad_before[2], pad_after[2]))

                data_t = np.pad(data[field][i], pad_vec, mode='constant')

                crop_amount = data_t.shape - self.size

                crop_before = np.floor(crop_amount / 2.).astype(int)
                crop_after = np.ceil(crop_amount / 2.).astype(int)

                crop_before[crop_before < 0] = 0
                crop_after[crop_after < 0] = 0

                crop_vec = (
                    (crop_before[0], crop_after[0]),
                    (crop_before[1], crop_after[1]),
                    (crop_before[2], crop_after[2])
                )

                data_t = data_t[
                         crop_vec[0][0]:data_t.shape[0] - crop_vec[0][1],
                         crop_vec[1][0]:data_t.shape[1] - crop_vec[1][1],
                         crop_vec[2][0]:data_t.shape[2] - crop_vec[2][1]
                         ]

                crops[field].append(crop_vec)
                pads[field].append(pad_vec)

                assert np.all(data_t.shape == self.size)

                data[field][i] = data_t

        data[self.field_pads] = pads
        data[self.field_crops] = crops
        data[self.field_original_sizes] = original_sizes
        return data


class FromNumpyStandardSizeToOriginalSize(object):
    def __init__(self,
                 fields,
                 field_pads='pads_std_size',
                 field_crops="crops_std_size",
                 field_original_sizes='original_sizes_std_size'
                 ):
        '''
        FromNumpyOriginalSizeToStandardSize resizes data to original size. This method pads with zeros
        therefore it does not restore any lost information due to cropping
        :param fields: fields of the dictionary whose content should be modified
        :param field_pads: field of data dictionary to use to read paddings used in this transform
        :param field_crops: field of data dictionary to use to read crops used in this transform
        :param field_original_sizes: field of data dictionary to use to read sizes used in this transform
        '''
        self.fields = fields

        self.field_pads = field_pads
        self.field_crops = field_crops
        self.field_original_sizes = field_original_sizes

    def __call__(self, data):
        pads = data[self.field_pads]
        crops = data[self.field_crops]
        original_sizes = data[self.field_original_sizes]
        for field in self.fields:
            for i in range(len(data[field])):
                data_t = np.pad(data[field][i], crops[field][i], mode='constant')
                data_t = data_t[
                         pads[field][i][0][0]:data_t.shape[0] - pads[field][i][0][1],
                         pads[field][i][1][0]:data_t.shape[1] - pads[field][i][1][1],
                         pads[field][i][2][0]:data_t.shape[2] - pads[field][i][2][1]
                         ]

                data[field][i] = data_t

                assert np.all(np.asarray(data[field][i].shape) == np.asarray(original_sizes[field][i]))

        return data


class FromListToNumpy5DArray(object):
    def __init__(self, fields):
        '''
        FromListToNumpy5DArray makes the data 5D and ready for inference
        :param fields: the fields that need to be transformed to 5D
        '''

        self.fields = fields

    def __call__(self, data):
        for field in self.fields:
            data[field] = np.stack(data[field])[..., np.newaxis]

        return data


class FromNumpy5DArrayToList(object):
    def __init__(self, fields):
        '''
        FromNumpy5DArrayToList makes the data a list of numpy tensors
        :param fields: the fields that need to be transformed to 5D
        '''

        self.fields = fields

    def __call__(self, data):
        for field in self.fields:
            data_t = []
            for i in range(data[field].shape[0]):
                data_t.append(np.squeeze(data[field][i]))

            data[field] = data_t
        return data


class FromLabelVolumeToVTKMesh(object):
    mesh_reduction_percentage = 0.90

    def __init__(
            self,
            label_filed,
            mesh_field,
            return_VTK_field='return_VTK',
            spacings_field='original_spacings_NP',
            sizes_field='original_sizes_std_size',
            origins_field='original_origins_NP',
            directions_field='original_directions_NP',
            convert_to_ras_field='RAS'
    ):

        self.label_filed = label_filed
        self.mesh_field = mesh_field
        self.return_VTK_field = return_VTK_field
        self.spacings_field = spacings_field
        self.sizes_field = sizes_field
        self.origins_field = origins_field
        self.directions_field = directions_field
        self.convert_to_ras_field = convert_to_ras_field

    def __call__(self, data):
        import vtk
        from vtk.util.numpy_support import numpy_to_vtk

        meshes = []

        for label, spacing, origin, size, direction, return_VTK, convert_to_ras in zip(
                data[self.label_filed],
                data[self.spacings_field][self.label_filed],
                data[self.origins_field][self.label_filed],
                data[self.sizes_field][self.label_filed],
                data[self.directions_field][self.label_filed],
                data[self.return_VTK_field],
                data[self.convert_to_ras_field],
        ):
            if return_VTK == "False":
                continue

            label = np.transpose(np.copy(label), [2, 1, 0])

            vtk_voxelmap = vtk.vtkImageData()
            vtk_voxelmap.SetSpacing(spacing[0], spacing[1], spacing[2])
            vtk_voxelmap.SetOrigin(origin[0], origin[1], origin[2])
            vtk_voxelmap.SetDimensions(size[0], size[1], size[2])
            vtk_voxelmap.SetScalarType(vtk.VTK_FLOAT)

            vtk_data_array = numpy_to_vtk(num_array=label.ravel(), deep=True, array_type=vtk.VTK_FLOAT)

            points = vtk_voxelmap.GetPointData()
            points.SetScalars(vtk_data_array)

            contour = vtk.vtkDiscreteMarchingCubes()
            contour.SetInput(vtk_voxelmap)
            contour.ComputeNormalsOn()
            contour.GenerateValues(1, 1, 1)
            contour.Update()

            polydata = vtk.vtkPolyData()
            polydata.DeepCopy(contour.GetOutput())

            smoother = vtk.vtkSmoothPolyDataFilter()
            smoother.SetInput(polydata)
            smoother.SetNumberOfIterations(150)
            smoother.SetFeatureAngle(60)
            smoother.SetRelaxationFactor(0.05)
            smoother.FeatureEdgeSmoothingOff()
            smoother.Update()

            transform = vtk.vtkTransform()
            if convert_to_ras == 'True':
                mat = np.asarray(direction).reshape([3, 3])
                conversion_mat = np.asarray([-1, 0, 0, 0, -1, 0, 0, 0, 1]).reshape([3, 3])
                direction = np.matmul(conversion_mat, mat).flatten()

            dir_homo = np.asarray([direction[0], direction[1], direction[2], 0.,
                                   direction[3], direction[4], direction[5], 0.,
                                   direction[6], direction[7], direction[8], 0.,
                                   0., 0., 0., 1.]).reshape([4, 4])

            transform.SetMatrix(dir_homo.flatten())

            transformFilter = vtk.vtkTransformPolyDataFilter()
            transformFilter.SetTransform(transform)
            transformFilter.SetInputConnection(smoother.GetOutputPort())
            transformFilter.Update()

            mesh = vtk.vtkPolyData()

            mesh.ShallowCopy(transformFilter.GetOutput())

            decimator = vtk.vtkDecimatePro()
            decimator.SetInput(mesh)
            decimator.SetTargetReduction(self.mesh_reduction_percentage)
            decimator.SetPreserveTopology(1)
            decimator.Update()

            clean_poly_data_filter = vtk.vtkCleanPolyData()
            clean_poly_data_filter.SetInput(decimator.GetOutput())
            clean_poly_data_filter.Update()

            meshes.append(clean_poly_data_filter.GetOutput())

        data[self.mesh_field] = meshes

        return data


class ThresholdNumpy(object):
    def __init__(self, image_field, threshold_field):
        self.image_field = image_field
        self.threshold_field = threshold_field

    def __call__(self, data):
        data[self.image_field] = (data[self.image_field] >= data[self.threshold_field]).astype(np.float32)

        return data
